@@ -1,9 +1,8 @@
-{-# LANGUAGE AutoDeriveTypeable  #-}
 {-# LANGUAGE BangPatterns        #-}
 {-# LANGUAGE CPP                 #-}
 {-# LANGUAGE DefaultSignatures   #-}
+{-# LANGUAGE DeriveDataTypeable  #-}
 {-# LANGUAGE DeriveGeneric       #-}
-{-# LANGUAGE LambdaCase          #-}
 {-# LANGUAGE OverloadedStrings   #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 
@@ -17,7 +16,63 @@
 -- writing adaptors to this library’s @'Doc'@ type, see
 -- "Data.Text.Prettyprint.Doc.Internal.Type".
 module Data.Text.Prettyprint.Doc.Internal (
-    module Data.Text.Prettyprint.Doc.Internal
+    -- * Documents
+    Doc(..),
+
+    -- * Basic functionality
+    Pretty(..),
+    viaShow, unsafeViaShow, unsafeTextWithoutNewlines,
+    emptyDoc, nest, line, line', softline, softline', hardline, group, flatAlt,
+
+    -- * Alignment functions
+    align, hang, indent, encloseSep, list, tupled,
+
+    -- * Binary functions
+    (<+>),
+
+    -- * List functions
+    concatWith,
+
+    -- ** 'sep' family
+    hsep, vsep, fillSep, sep,
+    -- ** 'cat' family
+    hcat, vcat, fillCat, cat,
+    -- ** Others
+    punctuate,
+
+    -- * Reactive/conditional layouts
+    column, nesting, width, pageWidth,
+
+    -- * Filler functions
+    fill, fillBreak,
+
+    -- * General convenience
+    plural, enclose, surround,
+
+    -- ** Annotations
+    annotate,
+    unAnnotate,
+    reAnnotate,
+    alterAnnotations,
+    unAnnotateS,
+    reAnnotateS,
+    alterAnnotationsS,
+
+    -- * Optimization
+    fuse, FusionDepth(..),
+
+    -- * Layout
+    SimpleDocStream(..),
+    PageWidth(..), defaultPageWidth,
+    LayoutOptions(..), defaultLayoutOptions,
+    layoutPretty, layoutCompact, layoutSmart,
+    removeTrailingWhitespace,
+
+    -- * Rendering
+    renderShowS,
+
+    -- * Internal helpers
+    textSpaces
 ) where
 
 
@@ -30,6 +85,7 @@ import           Data.String         (IsString (..))
 import           Data.Text           (Text)
 import qualified Data.Text           as T
 import qualified Data.Text.Lazy      as Lazy
+import           Data.Typeable       (Typeable)
 import           Data.Void
 import           Data.Word
 import           GHC.Generics        (Generic)
@@ -48,10 +104,6 @@ import Data.Traversable (Traversable (..))
 import Prelude          hiding (foldr, foldr1)
 #endif
 
-#if !(MONOID_IN_PRELUDE)
-import Data.Monoid hiding ((<>))
-#endif
-
 #if FUNCTOR_IDENTITY_IN_BASE
 import Data.Functor.Identity
 #endif
@@ -132,7 +184,7 @@ data Doc ann =
     -- | Add an annotation to the enclosed 'Doc'. Can be used for example to add
     -- styling directives or alt texts that can then be used by the renderer.
     | Annotated ann (Doc ann)
-    deriving (Generic)
+    deriving (Generic, Typeable)
 
 -- |
 -- @
@@ -197,10 +249,24 @@ class Pretty a where
     -- >>> prettyList [1, 23, 456]
     -- [1, 23, 456]
     prettyList :: [a] -> Doc ann
-    prettyList = list . map pretty
+    prettyList = align . list . map pretty
 
     {-# MINIMAL pretty #-}
 
+-- $
+-- Issue #67: Nested lists were not aligned with »pretty«, leading to non-pretty
+-- output, violating the Pretty class law.
+--
+-- >>> pretty (replicate 2 (replicate 4 (1, replicate 8 2)))
+-- [ [ (1, [2, 2, 2, 2, 2, 2, 2, 2])
+--   , (1, [2, 2, 2, 2, 2, 2, 2, 2])
+--   , (1, [2, 2, 2, 2, 2, 2, 2, 2])
+--   , (1, [2, 2, 2, 2, 2, 2, 2, 2]) ]
+-- , [ (1, [2, 2, 2, 2, 2, 2, 2, 2])
+--   , (1, [2, 2, 2, 2, 2, 2, 2, 2])
+--   , (1, [2, 2, 2, 2, 2, 2, 2, 2])
+--   , (1, [2, 2, 2, 2, 2, 2, 2, 2]) ] ]
+
 instance Pretty a => Pretty (Const a b) where
   pretty = pretty . getConst
 
@@ -232,9 +298,8 @@ instance Pretty () where
 -- | >>> pretty True
 -- True
 instance Pretty Bool where
-    pretty = \case
-        True  -> "True"
-        False -> "False"
+    pretty True  = "True"
+    pretty False = "False"
 
 -- | Instead of @('pretty' '\n')@, consider using @'line'@ as a more readable
 -- alternative.
@@ -286,7 +351,7 @@ instance Pretty Natural where pretty = unsafeViaShow
 instance Pretty Float where pretty = unsafeViaShow
 
 -- | >>> pretty (exp 1 :: Double)
--- 2.718281828459045
+-- 2.71828182845904...
 instance Pretty Double where pretty = unsafeViaShow
 
 -- | >>> pretty (123, "hello")
@@ -394,9 +459,9 @@ unsafeTextWithoutNewlines text = case T.uncons text of
 emptyDoc :: Doc ann
 emptyDoc = Empty
 
--- | @('nest' i x)@ lays out the document @x@ with the current indentation level
--- increased by @i@. Negative values are allowed, and decrease the nesting level
--- accordingly.
+-- | @('nest' i x)@ lays out the document @x@ with the current nesting level
+-- (indentation of the following lines) increased by @i@. Negative values are
+-- allowed, and decrease the nesting level accordingly.
 --
 -- >>> vsep [nest 4 (vsep ["lorem", "ipsum", "dolor"]), "sit", "amet"]
 -- lorem
@@ -405,7 +470,12 @@ emptyDoc = Empty
 -- sit
 -- amet
 --
--- See also 'hang', 'align' and 'indent'.
+-- See also
+--
+--   * 'hang' ('nest' relative to current cursor position instead of
+--      current nesting level)
+--   * 'align' (set nesting level to current cursor position)
+--   * 'indent' (increase indentation on the spot, padding with spaces).
 nest
     :: Int -- ^ Change of nesting level
     -> Doc ann
@@ -459,7 +529,7 @@ line' = FlatAlt Line mempty
 -- 'softline' = 'group' 'line'
 -- @
 softline :: Doc ann
-softline = group line
+softline = Union (Char ' ') Line
 
 -- | @'softline''@ is like @'softline'@, but behaves like @'mempty'@ if the
 -- resulting output does not fit on the page (instead of @'space'@). In other
@@ -481,7 +551,7 @@ softline = group line
 -- 'softline'' = 'group' 'line''
 -- @
 softline' :: Doc ann
-softline' = group line'
+softline' = Union mempty Line
 
 -- | A @'hardline'@ is /always/ laid out as a line break, even when 'group'ed or
 -- when there is plenty of space. Note that it might still be simply discarded
@@ -507,9 +577,16 @@ hardline = Line
 -- use of it.
 group :: Doc ann -> Doc ann
 -- See note [Group: special flattening]
-group x = case changesUponFlattening x of
-    Nothing -> x
-    Just x' -> Union x' x
+group x = case x of
+    Union{} -> x
+    FlatAlt a b -> case changesUponFlattening b of
+        Flattened b' -> Union b' a
+        AlreadyFlat  -> Union b a
+        NeverFlat    -> a
+    _ -> case changesUponFlattening x of
+        Flattened x' -> Union x' x
+        AlreadyFlat  -> x
+        NeverFlat    -> x
 
 -- Note [Group: special flattening]
 --
@@ -524,39 +601,56 @@ group x = case changesUponFlattening x of
 -- See https://github.com/quchen/prettyprinter/issues/22 for the  corresponding
 -- ticket.
 
+data FlattenResult a
+    = Flattened a
+    -- ^ @a@ is likely flatter than the input.
+    | AlreadyFlat
+    -- ^ The input was already flat, e.g. a 'Text'.
+    | NeverFlat
+    -- ^ The input couldn't be flattened: It contained a 'Line' or 'Fail'.
+
+instance Functor FlattenResult where
+    fmap f (Flattened a) = Flattened (f a)
+    fmap _ AlreadyFlat   = AlreadyFlat
+    fmap _ NeverFlat     = NeverFlat
+
 -- | Choose the first element of each @Union@, and discard the first field of
 -- all @FlatAlt@s.
 --
--- The result is 'Just' if the element might change depending on the layout
--- algorithm (i.e. contains differently renderable sub-documents), and 'Nothing'
--- if the document is static (e.g. contains only a plain 'Empty' node). See
--- [Group: special flattening] for further explanations.
-changesUponFlattening :: Doc ann -> Maybe (Doc ann)
-changesUponFlattening = \case
-    FlatAlt _ y     -> Just (flatten y)
-    Line            -> Just Fail
-    Union x _       -> changesUponFlattening x <|> Just x
+-- The result is 'Flattened' if the element might change depending on the layout
+-- algorithm (i.e. contains differently renderable sub-documents), and 'AlreadyFlat'
+-- if the document is static (e.g. contains only a plain 'Empty' node).
+-- 'NeverFlat' is returned when the document cannot be flattened because it
+-- contains a hard 'Line' or 'Fail'.
+-- See [Group: special flattening] for further explanations.
+changesUponFlattening :: Doc ann -> FlattenResult (Doc ann)
+changesUponFlattening = \doc -> case doc of
+    FlatAlt _ y     -> Flattened (flatten y)
+    Line            -> NeverFlat
+    Union x _       -> Flattened x
     Nest i x        -> fmap (Nest i) (changesUponFlattening x)
     Annotated ann x -> fmap (Annotated ann) (changesUponFlattening x)
 
-    Column f        -> Just (Column (flatten . f))
-    Nesting f       -> Just (Nesting (flatten . f))
-    WithPageWidth f -> Just (WithPageWidth (flatten . f))
+    Column f        -> Flattened (Column (flatten . f))
+    Nesting f       -> Flattened (Nesting (flatten . f))
+    WithPageWidth f -> Flattened (WithPageWidth (flatten . f))
 
     Cat x y -> case (changesUponFlattening x, changesUponFlattening y) of
-        (Nothing, Nothing) -> Nothing
-        (Just x', Nothing) -> Just (Cat x' y )
-        (Nothing, Just y') -> Just (Cat x  y')
-        (Just x', Just y') -> Just (Cat x' y')
+        (NeverFlat    ,  _          ) -> NeverFlat
+        (_            , NeverFlat   ) -> NeverFlat
+        (Flattened x' , Flattened y') -> Flattened (Cat x' y')
+        (Flattened x' , AlreadyFlat ) -> Flattened (Cat x' y)
+        (AlreadyFlat  , Flattened y') -> Flattened (Cat x y')
+        (AlreadyFlat  , AlreadyFlat ) -> AlreadyFlat
 
-    Empty  -> Nothing
-    Char{} -> Nothing
-    Text{} -> Nothing
-    Fail   -> Nothing
+    Empty  -> AlreadyFlat
+    Char{} -> AlreadyFlat
+    Text{} -> AlreadyFlat
+    Fail   -> NeverFlat
   where
     -- Flatten, but don’t report whether anything changes.
     flatten :: Doc ann -> Doc ann
-    flatten = \case
+    flatten = \doc -> case doc of
         FlatAlt _ y     -> flatten y
         Cat x y         -> Cat (flatten x) (flatten y)
         Nest i x        -> Nest i (flatten x)
@@ -764,6 +858,7 @@ tupled = group . encloseSep (flatAlt "( " "(")
 -- @
 (<+>) :: Doc ann -> Doc ann -> Doc ann
 x <+> y = x <> Char ' ' <> y
+infixr 6 <+> -- like <>
 
 
 
@@ -1101,7 +1196,10 @@ fillBreak f x = width x (\w ->
 
 -- | Insert a number of spaces. Negative values count as 0.
 spaces :: Int -> Doc ann
-spaces n = unsafeTextWithoutNewlines (T.replicate n " ")
+spaces n
+  | n <= 0    = Empty
+  | n == 1    = Char ' '
+  | otherwise = Text n (textSpaces n)
 
 -- $
 -- prop> \(NonNegative n) -> length (show (spaces n)) == n
@@ -1230,7 +1328,7 @@ reAnnotate re = alterAnnotations (pure . re)
 alterAnnotations :: (ann -> [ann']) -> Doc ann -> Doc ann'
 alterAnnotations re = go
   where
-    go = \case
+    go = \doc -> case doc of
         Fail     -> Fail
         Empty    -> Empty
         Char c   -> Char c
@@ -1256,7 +1354,7 @@ alterAnnotations re = go
 unAnnotateS :: SimpleDocStream ann -> SimpleDocStream xxx
 unAnnotateS = go
   where
-    go = \case
+    go = \doc -> case doc of
         SFail              -> SFail
         SEmpty             -> SEmpty
         SChar c rest       -> SChar c (go rest)
@@ -1269,7 +1367,7 @@ unAnnotateS = go
 reAnnotateS :: (ann -> ann') -> SimpleDocStream ann -> SimpleDocStream ann'
 reAnnotateS re = go
   where
-    go = \case
+    go = \doc -> case doc of
         SFail             -> SFail
         SEmpty            -> SEmpty
         SChar c rest      -> SChar c (go rest)
@@ -1279,6 +1377,7 @@ reAnnotateS re = go
         SAnnPush ann rest -> SAnnPush (re ann) (go rest)
 
 data AnnotationRemoval = Remove | DontRemove
+  deriving Typeable
 
 -- | Change the annotation of a document to a different annotation, or none at
 -- all. 'alterAnnotations' for 'SimpleDocStream'.
@@ -1292,7 +1391,7 @@ alterAnnotationsS re = go []
   where
     -- We keep a stack of whether to remove a pop so that we can remove exactly
     -- the pops corresponding to annotations that mapped to Nothing.
-    go stack = \case
+    go stack = \sds -> case sds of
         SFail             -> SFail
         SEmpty            -> SEmpty
         SChar c rest      -> SChar c (go stack rest)
@@ -1323,7 +1422,7 @@ data FusionDepth =
     -- This value should only be used if profiling shows it is significantly
     -- faster than using 'Shallow'.
     | Deep
-    deriving (Eq, Ord, Show)
+    deriving (Eq, Ord, Show, Typeable)
 
 -- | @('fuse' depth doc)@ combines text nodes so they can be rendered more
 -- efficiently. A fused document is always laid out identical to its unfused
@@ -1355,7 +1454,7 @@ data FusionDepth =
 fuse :: FusionDepth -> Doc ann -> Doc ann
 fuse depth = go
   where
-    go = \case
+    go = \doc -> case doc of
         Cat Empty x                   -> go x
         Cat x Empty                   -> go x
         Cat (Char c1) (Char c2)       -> Text 2 (T.singleton c1 <> T.singleton c2)
@@ -1381,7 +1480,7 @@ fuse depth = go
         Nest 0 x         -> go x
         Nest i x         -> Nest i (go x)
 
-        Annotated _ Empty -> Empty
+        Annotated ann x -> Annotated ann (go x)
 
         FlatAlt x1 x2 -> FlatAlt (go x1) (go x2)
         Union x1 x2   -> Union (go x1) (go x2)
@@ -1426,7 +1525,7 @@ data SimpleDocStream ann =
 
     -- | Remove a previously pushed annotation.
     | SAnnPop (SimpleDocStream ann)
-    deriving (Eq, Ord, Show, Generic)
+    deriving (Eq, Ord, Show, Generic, Typeable)
 
 -- | Remove all trailing space characters.
 --
@@ -1438,51 +1537,86 @@ data SimpleDocStream ann =
 -- whitespace, for example a renderer that colors the background of trailing
 -- whitespace, as e.g. @git diff@ can be configured to do.
 removeTrailingWhitespace :: SimpleDocStream ann -> SimpleDocStream ann
-removeTrailingWhitespace = go (WssWithheldWhitespace Nothing 0)
+removeTrailingWhitespace = go (RecordedWhitespace [] 0)
   where
-    commitSpaces :: Maybe Int -> Int -> SimpleDocStream ann -> SimpleDocStream ann
-    commitSpaces mWithheldNewline withheldSpaces = nl . sp
-      where
-        nl = case mWithheldNewline of
-            Just withheldLine -> SLine withheldLine
-            Nothing -> id
-        sp = case withheldSpaces of
-            0 -> id
-            1 -> SChar ' '
-            n -> SText n (T.replicate n " ")
+    commitWhitespace
+        :: [Int] -- Withheld lines
+        -> Int -- Withheld spaces
+        -> SimpleDocStream ann
+        -> SimpleDocStream ann
+    commitWhitespace is !n sds = case is of
+        []      -> case n of
+                       0 -> sds
+                       1 -> SChar ' ' sds
+                       _ -> SText n (textSpaces n) sds
+        (i:is') -> let !end = SLine (i + n) sds
+                   in prependEmptyLines is' end
+
+    prependEmptyLines :: [Int] -> SimpleDocStream ann -> SimpleDocStream ann
+    prependEmptyLines is sds0 = foldr (\_ sds -> SLine 0 sds) sds0 is
 
     go :: WhitespaceStrippingState -> SimpleDocStream ann -> SimpleDocStream ann
-    go annLevel@(WssAnnotationLevel annotationLevel) = \case
-        SFail -> SFail
-        SEmpty -> SEmpty
-        SChar c rest -> SChar c (go annLevel rest)
-        SText textLength text rest -> SText textLength text (go annLevel rest)
-        SLine i rest -> SLine i (go annLevel rest)
-        SAnnPush ann rest -> SAnnPush ann (go (WssAnnotationLevel (annotationLevel+1)) rest)
+    -- We do not strip whitespace inside annotated documents, since it might
+    -- actually be relevant there.
+    go annLevel@(AnnotationLevel annLvl) = \sds -> case sds of
+        SFail             -> SFail
+        SEmpty            -> SEmpty
+        SChar c rest      -> SChar c (go annLevel rest)
+        SText l text rest -> SText l text (go annLevel rest)
+        SLine i rest      -> SLine i (go annLevel rest)
+        SAnnPush ann rest -> let !annLvl' = annLvl+1
+                             in SAnnPush ann (go (AnnotationLevel annLvl') rest)
         SAnnPop rest
-            | annotationLevel > 1 -> SAnnPop (go (WssAnnotationLevel (annotationLevel-1)) rest)
-            | otherwise -> SAnnPop (go (WssWithheldWhitespace Nothing 0) rest)
-    go (WssWithheldWhitespace withheldLine withheldSpaces) = \case
+            | annLvl > 1  -> let !annLvl' = annLvl-1
+                             in SAnnPop (go (AnnotationLevel annLvl') rest)
+            | otherwise   -> SAnnPop (go (RecordedWhitespace [] 0) rest)
+    -- Record all spaces/lines encountered, and once proper text starts again,
+    -- release only the necessary ones.
+    go (RecordedWhitespace withheldLines withheldSpaces) = \sds -> case sds of
         SFail -> SFail
-        SEmpty -> SEmpty
+        SEmpty -> prependEmptyLines withheldLines SEmpty
         SChar c rest
-            | c == ' ' -> go (WssWithheldWhitespace withheldLine (withheldSpaces+1)) rest
-            | otherwise -> commitSpaces withheldLine withheldSpaces (SChar c (go (WssWithheldWhitespace Nothing 0) rest))
+            | c == ' ' -> go (RecordedWhitespace withheldLines (withheldSpaces+1)) rest
+            | otherwise -> commitWhitespace
+                               withheldLines
+                               withheldSpaces
+                               (SChar c (go (RecordedWhitespace [] 0) rest))
         SText textLength text rest ->
             let stripped = T.dropWhileEnd (== ' ') text
                 strippedLength = T.length stripped
                 trailingLength = textLength - strippedLength
                 isOnlySpace = strippedLength == 0
             in if isOnlySpace
-                then go (WssWithheldWhitespace withheldLine (withheldSpaces + textLength)) rest
-                else commitSpaces withheldLine withheldSpaces (SText strippedLength stripped (go (WssWithheldWhitespace Nothing trailingLength) rest))
-        SLine i rest -> go (WssWithheldWhitespace (Just i) 0) rest
-        SAnnPush ann rest -> commitSpaces withheldLine withheldSpaces (SAnnPush ann (go (WssAnnotationLevel 1) rest))
+                then go (RecordedWhitespace withheldLines (withheldSpaces + textLength)) rest
+                else commitWhitespace
+                        withheldLines
+                        withheldSpaces
+                        (SText strippedLength
+                               stripped
+                               (go (RecordedWhitespace [] trailingLength) rest))
+        SLine i rest -> go (RecordedWhitespace (i:withheldLines) 0) rest
+        SAnnPush ann rest -> commitWhitespace
+                                 withheldLines
+                                 withheldSpaces
+                                 (SAnnPush ann (go (AnnotationLevel 1) rest))
         SAnnPop _ -> error "Tried skipping spaces in unannotated data! Please report this as a bug in 'prettyprinter'."
 
 data WhitespaceStrippingState
-    = WssAnnotationLevel !Int
-    | WssWithheldWhitespace (Maybe Int) !Int
+    = AnnotationLevel !Int
+    | RecordedWhitespace [Int] !Int
+      -- ^ [Newline with indentation i] Spaces
+  deriving Typeable
+
+
+
+-- $
+-- >>> import qualified Data.Text.IO as T
+-- >>> doc = "lorem" <> hardline <> hardline <> pretty "ipsum"
+-- >>> go = T.putStrLn . renderStrict . removeTrailingWhitespace . layoutPretty defaultLayoutOptions
+-- >>> go doc
+-- lorem
+-- <BLANKLINE>
+-- ipsum
 
 
 
@@ -1499,7 +1633,7 @@ instance Functor SimpleDocStream where
 instance Foldable SimpleDocStream where
     foldMap f = go
       where
-        go = \case
+        go = \sds -> case sds of
             SFail             -> mempty
             SEmpty            -> mempty
             SChar _ rest      -> go rest
@@ -1513,7 +1647,7 @@ instance Foldable SimpleDocStream where
 instance Traversable SimpleDocStream where
     traverse f = go
       where
-        go = \case
+        go = \sds -> case sds of
             SFail             -> pure SFail
             SEmpty            -> pure SEmpty
             SChar c rest      -> SChar c   <$> go rest
@@ -1526,19 +1660,21 @@ instance Traversable SimpleDocStream where
 --
 --   - page width
 --   - minimum nesting level to fit in
---   - width in which to fit the first line; Nothing is unbounded
+--   - width in which to fit the first line
 newtype FittingPredicate ann
   = FittingPredicate (PageWidth
                    -> Int
-                   -> Maybe Int
+                   -> Int
                    -> SimpleDocStream ann
                    -> Bool)
+  deriving Typeable
 
 -- | List of nesting level/document pairs yet to be laid out.
 data LayoutPipeline ann =
       Nil
     | Cons !Int (Doc ann) (LayoutPipeline ann)
     | UndoAnn (LayoutPipeline ann)
+  deriving Typeable
 
 -- | Maximum number of characters that fit in one line. The layout algorithms
 -- will try not to exceed the set limit by inserting line breaks when applicable
@@ -1559,7 +1695,10 @@ data PageWidth
     | Unbounded
     -- ^ Layouters should not introduce line breaks on their own.
 
-    deriving (Eq, Ord, Show)
+    deriving (Eq, Ord, Show, Typeable)
+
+defaultPageWidth :: PageWidth
+defaultPageWidth = AvailablePerLine 80 1
 
 -- $ Test to avoid surprising behaviour
 -- >>> Unbounded > AvailablePerLine maxBound 1
@@ -1567,7 +1706,7 @@ data PageWidth
 
 -- | Options to influence the layout algorithms.
 newtype LayoutOptions = LayoutOptions { layoutPageWidth :: PageWidth }
-    deriving (Eq, Ord, Show)
+    deriving (Eq, Ord, Show, Typeable)
 
 -- | The default layout options, suitable when you just want some output, and
 -- don’t particularly care about the details. Used by the 'Show' instance, for
@@ -1576,7 +1715,7 @@ newtype LayoutOptions = LayoutOptions { layoutPageWidth :: PageWidth }
 -- >>> defaultLayoutOptions
 -- LayoutOptions {layoutPageWidth = AvailablePerLine 80 1.0}
 defaultLayoutOptions :: LayoutOptions
-defaultLayoutOptions = LayoutOptions { layoutPageWidth = AvailablePerLine 80 1 }
+defaultLayoutOptions = LayoutOptions { layoutPageWidth = defaultPageWidth }
 
 -- | This is the default layout algorithm, and it is used by 'show', 'putDoc'
 -- and 'hPutDoc'.
@@ -1591,9 +1730,7 @@ layoutPretty
     -> Doc ann
     -> SimpleDocStream ann
 layoutPretty = layoutWadlerLeijen
-    (FittingPredicate (\_pWidth _minNestingLevel maxWidth sdoc -> case maxWidth of
-        Nothing -> True
-        Just w -> fits w sdoc ))
+    (FittingPredicate (\_pWidth _minNestingLevel maxWidth sdoc -> fits maxWidth sdoc))
   where
     fits :: Int -- ^ Width in which to fit the first line
          -> SimpleDocStream ann
@@ -1611,7 +1748,7 @@ layoutPretty = layoutWadlerLeijen
 -- line breaks earlier if the content does not (or will not, rather) fit into
 -- one line.
 --
--- Considre the following python-ish document,
+-- Consider the following python-ish document,
 --
 -- >>> let fun x = hang 2 ("fun(" <> softline' <> x) <> ")"
 -- >>> let doc = (fun . fun . fun . fun . fun) (align (list ["abcdef", "ghijklm"]))
@@ -1624,7 +1761,7 @@ layoutPretty = layoutWadlerLeijen
 -- >>> let hr = pipe <> pretty (replicate (26-2) '-') <> pipe
 -- >>> let go layouter x = (T.putStrLn . renderStrict . layouter (LayoutOptions (AvailablePerLine 26 1))) (vsep [hr, x, hr])
 --
--- If we render this using @'layoutPretty'@ with a page width of 26 characters
+-- If we render this using 'layoutPretty' with a page width of 26 characters
 -- per line, all the @fun@ calls fit into the first line so they will be put
 -- there,
 --
@@ -1649,25 +1786,38 @@ layoutPretty = layoutWadlerLeijen
 --           , ghijklm ])))))
 -- |------------------------|
 --
--- The key difference between @'layoutPretty'@ and @'layoutSmart'@ is that the
--- latter will check the potential document up to the end of the current
--- indentation level, instead of just having one element lookahead.
+-- The key difference between 'layoutPretty' and 'layoutSmart' is that the
+-- latter will check the potential document until it encounters a line with the
+-- same indentation or less than the start of the document. Any line encountered
+-- earlier is assumed to belong to the same syntactic structure.
+-- 'layoutPretty' checks only the first line.
+--
+-- Consider for example the question of whether the @A@s fit into the document
+-- below:
+--
+-- > 1 A
+-- > 2   A
+-- > 3  A
+-- > 4 B
+-- > 5   B
+--
+-- 'layoutPretty' will check only line 1, ignoring whether e.g. line 2 might
+-- already be too wide.
+-- By contrast, 'layoutSmart' stops only once it reaches line 4, where the @B@
+-- has the same indentation as the first @A@.
 layoutSmart
     :: LayoutOptions
     -> Doc ann
     -> SimpleDocStream ann
-layoutSmart = layoutWadlerLeijen
-    (FittingPredicate (\pWidth minNestingLevel maxWidth sdoc -> case maxWidth of
-        Nothing -> True
-        Just w -> fits pWidth minNestingLevel w sdoc ))
+layoutSmart = layoutWadlerLeijen (FittingPredicate fits)
   where
-    -- Search with more lookahead: assuming that nesting roughly corresponds to
-    -- syntactic depth, @fits@ checks that not only the current line fits, but
-    -- the entire syntactic structure being formatted at this level of
-    -- indentation fits. If we were to remove the second case for @SLine@, we
-    -- would check that not only the current structure fits, but also the rest
-    -- of the document, which would be slightly more intelligent but would have
-    -- exponential runtime (and is prohibitively expensive in practice).
+    -- Why doesn't layoutSmart simply check the entire document?
+    --
+    -- 1. That would be very expensive.
+    -- 2. In that case the layout of a particular part of a document would
+    --    depend on the fit of completely unrelated parts of the same document.
+    --    See https://github.com/quchen/prettyprinter/issues/83 for a related
+    --    bug.
     fits :: PageWidth
          -> Int -- ^ Minimum nesting level to fit in
          -> Int -- ^ Width in which to fit the first line
@@ -1691,7 +1841,7 @@ layoutWadlerLeijen
     -> Doc ann
     -> SimpleDocStream ann
 layoutWadlerLeijen
-    fittingPredicate
+    (FittingPredicate fits)
     LayoutOptions { layoutPageWidth = pWidth }
     doc
   = best 0 0 (Cons 0 doc Nil)
@@ -1717,39 +1867,188 @@ layoutWadlerLeijen
         Nest j x        -> let !ij = i+j in best nl cc (Cons ij x ds)
         Union x y       -> let x' = best nl cc (Cons i x ds)
                                y' = best nl cc (Cons i y ds)
-                           in selectNicer fittingPredicate nl cc x' y'
+                           in selectNicer nl cc x' y'
         Column f        -> best nl cc (Cons i (f cc) ds)
         WithPageWidth f -> best nl cc (Cons i (f pWidth) ds)
         Nesting f       -> best nl cc (Cons i (f i) ds)
         Annotated ann x -> SAnnPush ann (best nl cc (Cons i x (UndoAnn ds)))
 
+    -- Select the better fitting of two documents:
+    -- Choice A if it fits, otherwise choice B.
+    --
+    -- The fit of choice B is /not/ checked! It is ultimately the user's
+    -- responsibility to provide an alternative that can fit the page even when
+    -- choice A doesn't.
     selectNicer
-        :: FittingPredicate ann
-        -> Int           -- ^ Current nesting level
+        :: Int           -- ^ Current nesting level
         -> Int           -- ^ Current column
-        -> SimpleDocStream ann -- ^ Choice A. Invariant: first lines should not be longer than B's.
-        -> SimpleDocStream ann -- ^ Choice B.
+        -> SimpleDocStream ann -- ^ Choice A.
+        -> SimpleDocStream ann -- ^ Choice B. Should fit more easily
+                               --   (== be less wide) than choice A.
         -> SimpleDocStream ann -- ^ Choice A if it fits, otherwise B.
-    selectNicer (FittingPredicate fits) lineIndent currentColumn x y
-      | fits pWidth minNestingLevel availableWidth x = x
-      | otherwise = y
+    selectNicer lineIndent currentColumn x y = case pWidth of
+        AvailablePerLine lineLength ribbonFraction
+          | fits pWidth minNestingLevel availableWidth x -> x
+          where
+            minNestingLevel =
+                -- See the Note
+                -- [Choosing the right minNestingLevel for consistent smart layouts]
+                case initialIndentation y of
+                    Just i ->
+                        -- y could be a (less wide) hanging layout. If so, let's
+                        -- check x a bit more thoroughly so we don't miss a potentially
+                        -- better fitting y.
+                        min i currentColumn
+                    Nothing ->
+                        -- y definitely isn't a hanging layout. Let's check x with the
+                        -- same minNestingLevel that any subsequent lines with the same
+                        -- indentation use.
+                        currentColumn
+            availableWidth = min columnsLeftInLine columnsLeftInRibbon
+              where
+                columnsLeftInLine = lineLength - currentColumn
+                columnsLeftInRibbon = lineIndent + ribbonWidth - currentColumn
+                ribbonWidth =
+                    (max 0 . min lineLength . round)
+                        (fromIntegral lineLength * ribbonFraction)
+        Unbounded
+          -- See the Note [Detecting failure with Unbounded page width].
+          | not (failsOnFirstLine x) -> x
+        _ -> y
+
+    initialIndentation :: SimpleDocStream ann -> Maybe Int
+    initialIndentation sds = case sds of
+        SLine i _    -> Just i
+        SAnnPush _ s -> initialIndentation s
+        SAnnPop s    -> initialIndentation s
+        _            -> Nothing
+
+    failsOnFirstLine :: SimpleDocStream ann -> Bool
+    failsOnFirstLine = go
       where
-        minNestingLevel = min lineIndent currentColumn
-        ribbonWidth = case pWidth of
-            AvailablePerLine lineLength ribbonFraction ->
-                (Just . max 0 . min lineLength . round)
-                    (fromIntegral lineLength * ribbonFraction)
-            Unbounded -> Nothing
-        availableWidth = do
-            columnsLeftInLine <- case pWidth of
-                AvailablePerLine cpl _ribbonFrac -> Just (cpl - currentColumn)
-                Unbounded -> Nothing
-            columnsLeftInRibbon <- do
-                li <- Just lineIndent
-                rw <- ribbonWidth
-                cc <- Just currentColumn
-                Just (li + rw - cc)
-            Just (min columnsLeftInLine columnsLeftInRibbon)
+        go sds = case sds of
+            SFail        -> True
+            SEmpty       -> False
+            SChar _ s    -> go s
+            SText _ _ s  -> go s
+            SLine _ _    -> False
+            SAnnPush _ s -> go s
+            SAnnPop s    -> go s
+
+
+{- Note [Choosing the right minNestingLevel for consistent smart layouts]
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+Consider this document:
+
+    doc =
+            "Groceries: "
+        <>  align
+                (cat
+                    [ sep ["pommes", "de", "terre"]
+                    , "apples"
+                    , "Donaudampfschifffahrtskapitänskajütenmülleimer"
+                    ]
+                )
+
+... and assume we want to fit it into 40 columns as nicely as possible:
+
+    opts = LayoutOptions (AvailablePerLine 40 1)
+
+We already have bad luck with the last item – it's longer than 40 characters
+on its own!
+
+We'd still like the first item, pommes de terre, to be laid out nicely, that is,
+on one line, since it's not too wide. This is what we'd like to see:
+
+    Groceries: pommes de terre
+               apples
+               Donaudampfschifffahrtskapitänskajütenmülleimer
+
+Before #83 was fixed, that wasn't what we got! Instead we got this:
+
+> renderIO stdout $ layoutSmart opts doc
+Groceries: pommes
+           de
+           terre
+           apples
+           Donaudampfschifffahrtskapitänskajütenmülleimer
+
+Why?
+
+minNestingLevel was effectively defined as
+
+    minNestingLevel = lineIndent
+
+The lineIndent for "pommes de terre" is 0.
+
+The FittingPredicate for layoutSmart will continue to check the rest of the
+document until it finds a line where the indentation <= minNestingLevel.
+In this case this meant that layoutSmart would traverse all the items,
+and note that the last item, Donaudampfschifffahrtskapitänskajütenmülleimer,
+doesn't fit into the available space! The "flatter" version of the document
+has failed, so "pommes de terre" gets spread over several lines!
+
+Obviously this would be an inconsistency with the layout of the other items.
+Their lineIndent is 11 each, so for them, the FittingPredicate stops already
+on the next line.
+
+The obvious solution is to change the definition of minNestingLevel:
+
+    minNestingLevel = currentColumn
+
+This however breaks the "python-ish" document from the documentation for
+layoutSmart:
+
+    expected: |------------------------|
+              fun(
+                fun(
+                  fun(
+                    fun(
+                      fun(
+                        [ abcdef
+                        , ghijklm ])))))
+              |------------------------|
+
+     but got: |------------------------|
+              fun(
+                fun(
+                  fun(
+                    fun(
+                      fun([ abcdef
+                          , ghijklm ])))))
+              |------------------------|
+
+We now accept the worse layout because the problematic last line has
+the same indentation as the current column of "[ abcdef", so we don't check it!
+
+The solution we went with in the end is a bit of a hack:
+
+We check whether the alternative, "high" layout is a (potentially less wide)
+hanging layout, and in that case pick its indentation as the minNestingLevel.
+
+This way we achieve the optimal layout in both scenarios.
+
+See https://github.com/quchen/prettyprinter/issues/83 for the bug that lead
+to the current solution.
+
+
+Note [Detecting failure with Unbounded page width]
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+To understand why it is sufficient to check the first line of the
+SimpleDocStream, trace how an SFail ends up there:
+
+1. We group a Doc containing a Line, producing a (Union x y) where
+   x contains Fail.
+
+2. In best, any Unions are handled recursively, rejecting any
+   alternatives that would result in SFail.
+
+So once a SimpleDocStream reaches selectNicer, any SFail in it must
+appear before the first linebreak – any other SFail would have been
+detected and rejected in a previous iteration.
+-}
+
+
 
 -- | @(layoutCompact x)@ lays out the document @x@ without adding any
 -- indentation. Since no \'pretty\' printing is involved, this layouter is very
@@ -1801,7 +2100,7 @@ instance Show (Doc ann) where
 --     'showsPrec' _ = 'renderShowS' . 'layoutPretty' 'defaultLayoutOptions' . 'pretty'
 -- @
 renderShowS :: SimpleDocStream ann -> ShowS
-renderShowS = \case
+renderShowS = \sds -> case sds of
     SFail        -> panicUncaughtFail
     SEmpty       -> id
     SChar c x    -> showChar c . renderShowS x
@@ -1811,6 +2110,19 @@ renderShowS = \case
     SAnnPop x    -> renderShowS x
 
 
+-- | A utility for producing indentation etc.
+--
+-- >>> textSpaces 3
+-- "   "
+--
+-- This produces much better Core than the equivalent
+--
+-- > T.replicate n " "
+--
+-- (See <https://github.com/quchen/prettyprinter/issues/131>.)
+textSpaces :: Int -> Text
+textSpaces n = T.replicate n (T.singleton ' ')
+
 
 -- $setup
 --
